package com.ds.ds.addressUtils;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.Stack;

import com.ds.ds.randomUtils.SRandom;
import org.bitcoinj.base.ScriptType;
import org.bitcoinj.crypto.ECKey;
import org.bitcoinj.params.MainNetParams;

class PAddress {
    public ECKey ec;
    public String address;
    public long timestamp;
    public long transactions_count;

    public PAddress(ECKey _ec, long _timestamp, long _t_num)
    {
        this.ec = _ec;
        this.timestamp = _timestamp;
        this.transactions_count = _t_num;
        this.address = this.ec.toAddress(ScriptType.P2WPKH, MainNetParams.get().network()).toString();
    }
}

class AddressChain
{
    public ArrayList<PAddress> addresses = new ArrayList<>();
    public ArrayList<Double> address_distro = new ArrayList<>();

    public void AddAddress(PAddress p, Double dist)
    {
        this.addresses.add(p);
        this.address_distro.add(dist);
    }

    public boolean IsValid()
    {
        if(this.addresses.size() == 0) return false;
        Double s = 0.0;
        for (int i = 0; i < this.address_distro.size(); i++) {
            s += this.address_distro.get(i);
        }
        return s > 0.99999999 && s <= 1.0000000;
    }
}

public class AddressPool {

    public static ArrayList<PAddress> pool = new ArrayList<>();

    static int MAX_ADDRESSES_IN_CHAIN = 10;

    public static void CreatePool(int num)
    {
        for (int i = 0; i < num; i++)
        {
            pool.add((new PAddress(new ECKey(), new Date().getTime(), 0)));
        }
    }

    public static Stack<AddressChain> CreateChain(int num)
    {
        Stack<AddressChain> t_stack = new Stack<>();
        int last_item = -1;
        int new_item = -1;
        int target_addresses = 0;
        for (int i = 0; i < num; i++)
        {
            AddressChain t_chain = new AddressChain();
            while (true)
            {
                double max = 1.0;
                target_addresses = (Math.abs(SRandom.GetRandomInt()) % MAX_ADDRESSES_IN_CHAIN) + 1;
                for (int j = 0; j < target_addresses; j++) {
                    while(last_item == new_item) {
                        new_item = Math.abs(SRandom.GetRandomInt()) % pool.size();
                    }
                    last_item = new_item;
                    double t;
                    if(j < target_addresses - 1)
                    {
                        double d = (Math.abs(SRandom.GetRandomInt()) % target_addresses)+1.136578+Math.sqrt(Math.abs(SRandom.GetRandomInt() % (target_addresses * 9.487993)));
                        t = max / d;
                        max -= t;
                    }
                    else
                    {
                        t = max;
                        max = 1.0;
                    }
                    t_chain.AddAddress(pool.get(new_item), t);
                }
                if(t_chain.IsValid())
                    break;
                else
                    t_chain = new AddressChain();
            }
            t_stack.push(t_chain);
        }
        return t_stack;
    }

    public static void PrintPool()
    {
        System.out.println("#####################Address Pool#####################");
        for (int i = 0; i < pool.size(); i++) {
            System.out.println("#" + i + ": " + pool.get(i).address);
            System.out.println("time: " + pool.get(i).timestamp + "; transactions: " + pool.get(i).transactions_count + ";");
        }
        System.out.println("######################################################");
    }

    public static void PrintChain(Stack<AddressChain> sss)
    {
        System.out.println("#####################Address Chain#####################");
        while (sss.size() > 0)
        {
            AddressChain ac = sss.pop();
            for (int i = 0; i < ac.addresses.size(); i++) {
                System.out.print(ac.addresses.get(i).address + ": " + ac.address_distro.get(i) * 100.0 + "%; ");
            }
            System.out.print("\n IsValid: " + ac.IsValid() + ";\n");
        }
        System.out.println("#######################################################");
    }

}
